@@ -1,5 +1,7 @@
+import sys
 import os
 import re
+import threading
 
 import zmq
 from zmq.utils import jsonapi
@@ -264,7 +266,8 @@ class Debugger:
 
     # Requests that can be handled even if the debugger is not running
     static_debug_msg_types = [
-        'debugInfo', 'inspectVariables', 'richInspectVariables'
+        'debugInfo', 'inspectVariables', 
+        'richInspectVariables', 'modules'
     ]
 
     def __init__(self, log, debugpy_stream, event_callback, shell_socket, session):
@@ -274,6 +277,7 @@ class Debugger:
         self.session = session
         self.is_started = False
         self.event_callback = event_callback
+        self.stopped_queue = Queue()
 
         self.started_debug_handlers = {}
         for msg_type in Debugger.started_debug_msg_types:
@@ -284,7 +288,7 @@ class Debugger:
             self.static_debug_handlers[msg_type] = getattr(self, msg_type)
 
         self.breakpoint_list = {}
-        self.stopped_threads = []
+        self.stopped_threads = set()
 
         self.debugpy_initialized = False
         self._removed_cleanup = {}
@@ -297,13 +301,19 @@ class Debugger:
 
     def _handle_event(self, msg):
         if msg['event'] == 'stopped':
-            self.stopped_threads.append(msg['body']['threadId'])
+            if msg['body']['allThreadsStopped']:
+                self.stopped_queue.put_nowait(msg)
+                # Do not forward the event now, will be done in the handle_stopped_event
+                return
+            else:
+                self.stopped_threads.add(msg['body']['threadId'])
+                self.event_callback(msg)
         elif msg['event'] == 'continued':
-            try:
+            if msg['body']['allThreadsContinued']:
+                self.stopped_threads = set()
+            else:
                 self.stopped_threads.remove(msg['body']['threadId'])
-            except Exception:
-                pass
-        self.event_callback(msg)
+            self.event_callback(msg)
 
     async def _forward_message(self, msg):
         return await self.debugpy_client.send_dap_request(msg)
@@ -322,6 +332,32 @@ class Debugger:
         }
         return reply
 
+    def _accept_stopped_thread(self, thread_name):
+        # TODO: identify Thread-2, Thread-3 and Thread-4. These are NOT
+        # Control, IOPub or Heartbeat threads
+        forbid_list = [
+            'IPythonHistorySavingThread',
+            'Thread-2',
+            'Thread-3',
+            'Thread-4'
+        ]
+        return thread_name not in forbid_list
+        
+    async def handle_stopped_event(self):
+        # Wait for a stopped event message in the stopped queue
+        # This message is used for triggering the 'threads' request
+        event = await self.stopped_queue.get()
+        req = {
+            'seq': event['seq'] + 1,
+            'type': 'request',
+            'command': 'threads'
+        }
+        rep = await self._forward_message(req)
+        for t in rep['body']['threads']:
+            if self._accept_stopped_thread(t['name']):
+                self.stopped_threads.add(t['id'])
+        self.event_callback(event)
+
     @property
     def tcp_client(self):
         return self.debugpy_client
@@ -513,7 +549,7 @@ class Debugger:
                 'tmpFilePrefix': get_tmp_directory() + os.sep,
                 'tmpFileSuffix': '.py',
                 'breakpoints': breakpoint_list,
-                'stoppedThreads': self.stopped_threads,
+                'stoppedThreads': list(self.stopped_threads),
                 'richRendering': True,
                 'exceptionPaths': ['Python Exceptions']
             }
@@ -581,6 +617,24 @@ class Debugger:
         reply["success"] = True
         return reply
 
+    async def modules(self, message):
+        modules = list(sys.modules.values())
+        startModule = message.get('startModule', 0)
+        moduleCount = message.get('moduleCount', len(modules))
+        mods = []
+        for i in range(startModule, moduleCount):
+            module = modules[i]
+            filename = getattr(getattr(module, '__spec__', None), 'origin', None)
+            if filename and filename.endswith('.py'):
+                mods.append({
+                    'id': i,
+                    'name': module.__name__,
+                    'path': filename
+                })
+
+        reply = {'body': {'modules': mods, 'totalModules': len(modules)}}
+        return reply
+
     async def process_request(self, message):
         reply = {}
 
@@ -613,7 +667,7 @@ class Debugger:
         if message['command'] == 'disconnect':
             self.stop()
             self.breakpoint_list = {}
-            self.stopped_threads = []
+            self.stopped_threads = set()
             self.is_started = False
             self.log.info('The debugger has stopped')
 
